@@ -12,7 +12,7 @@
 
  ****************************************************************************/
 
-#define BUCKETTEMPLATE_C "$Id: BucketTemplate.c 37245 2005-07-18 19:27:28Z tim_one $\n"
+#define BUCKETTEMPLATE_C "$Id: BucketTemplate.c 38291 2005-09-03 13:42:19Z alienoid $\n"
 
 /* Use BUCKET_SEARCH to find the index at which a key belongs.
  * INDEX    An int lvalue to hold the index i such that KEY belongs at
@@ -457,7 +457,12 @@ update_from_seq(PyObject *map, PyObject *seq)
        INCREF of the seq argument.  So seq must always be DECREFed on
        the way out.
      */
-    if (!PySequence_Check(seq)) {
+    /* Use items() if it's not a sequence.  Alas, PySequence_Check()
+     * returns true for a PeristentMapping or PersistentDict, and we
+     * want to use items() in those cases too.
+     */
+    if (!PySequence_Check(seq) || /* or it "looks like a dict" */
+          PyObject_HasAttrString(seq, "iteritems")) {
 	PyObject *items;
 	items = PyObject_GetAttrString(seq, "items");
 	if (items == NULL)
@@ -466,7 +471,8 @@ update_from_seq(PyObject *map, PyObject *seq)
 	Py_DECREF(items);
 	if (seq == NULL)
 	    return -1;
-    } else
+    }
+    else
 	Py_INCREF(seq);
 
     iter = PyObject_GetIter(seq);
@@ -1251,6 +1257,82 @@ bucket_has_key(Bucket *self, PyObject *key)
     return _bucket_get(self, key, 1);
 }
 
+static PyObject *
+bucket_setdefault(Bucket *self, PyObject *args)
+{
+    PyObject *key;
+    PyObject *failobj; /* default */
+    PyObject *value;   /* return value */
+    int dummy_changed; /* in order to call _bucket_set */
+
+    if (! PyArg_UnpackTuple(args, "setdefault", 2, 2, &key, &failobj))
+    	return NULL;
+
+    value = _bucket_get(self, key, 0);
+    if (value != NULL)
+        return value;
+
+    /* The key isn't in the bucket.  If that's not due to a KeyError exception,
+     * pass back the unexpected exception.
+     */
+    if (! PyErr_ExceptionMatches(PyExc_KeyError))
+        return NULL;
+    PyErr_Clear();
+
+    /* Associate `key` with `failobj` in the bucket, and return `failobj`. */
+    value = failobj;
+    if (_bucket_set(self, key, failobj, 0, 0, &dummy_changed) < 0)
+        value = NULL;
+    Py_XINCREF(value);
+    return value;
+}
+
+
+/* forward declaration */
+static int
+Bucket_length(Bucket *self);
+
+static PyObject *
+bucket_pop(Bucket *self, PyObject *args)
+{
+    PyObject *key;
+    PyObject *failobj = NULL; /* default */
+    PyObject *value;          /* return value */
+    int dummy_changed;        /* in order to call _bucket_set */
+
+    if (! PyArg_UnpackTuple(args, "pop", 1, 2, &key, &failobj))
+    	return NULL;
+
+    value = _bucket_get(self, key, 0);
+    if (value != NULL) {
+        /* Delete key and associated value. */
+        if (_bucket_set(self, key, NULL, 0, 0, &dummy_changed) < 0) {
+            Py_DECREF(value);
+            return NULL;
+        }
+        return value;
+    }
+
+    /* The key isn't in the bucket.  If that's not due to a KeyError exception,
+     * pass back the unexpected exception.
+     */
+    if (! PyErr_ExceptionMatches(PyExc_KeyError))
+        return NULL;
+
+    if (failobj != NULL) {
+    	/* Clear the KeyError and return the explicit default. */
+    	PyErr_Clear();
+    	Py_INCREF(failobj);
+    	return failobj;
+    }
+
+    /* No default given.  The only difference in this case is the error
+     * message, which depends on whether the bucket is empty.
+     */
+    if (Bucket_length(self) == 0)
+        PyErr_SetString(PyExc_KeyError, "pop(): Bucket is empty");
+    return NULL;
+ }
 
 /* Search bucket self for key.  This is the sq_contains slot of the
  * PySequenceMethods.
@@ -1475,6 +1557,16 @@ static struct PyMethodDef Bucket_methods[] = {
      "get(key[,default]) -- Look up a value\n\n"
      "Return the default (or None) if the key is not found."},
 
+    {"setdefault", (PyCFunction) bucket_setdefault, METH_VARARGS,
+     "D.setdefault(k, d) -> D.get(k, d), also set D[k]=d if k not in D.\n\n"
+     "Return the value like get() except that if key is missing, d is both\n"
+     "returned and inserted into the bucket as the value of k."},
+
+    {"pop", (PyCFunction) bucket_pop, METH_VARARGS,
+     "D.pop(k[, d]) -> v, remove key and return the corresponding value.\n\n"
+     "If key is not found, d is returned if given, otherwise KeyError\n"
+     "is raised."},
+
     {"iterkeys", (PyCFunction) Bucket_iterkeys,  METH_KEYWORDS,
      "B.iterkeys([min[,max]]) -> an iterator over the keys of B"},
 
